/*
  screen-uim - GNU screen filter with uim
  base on agent.c in uim-0.1.5

 $Id: screen-uim.c,v 1.4 2003/12/01 06:48:24 knok Exp $
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>

#include <uim/uim.h>
#include "udlib.h"

static struct agent_context {
  uim_context uc;
  int sfd;
  struct agent_context *next;
} default_context;

#define BUFSIZE 2048

char tickbuf[BUFSIZE];
char modebuf[BUFSIZE];

void settermraw(void);
void resetterm(void);
void sighandler(int);

static struct termios tbuf;

#define MAXLINE 2048
char *sockpath = NULL;

static void
commit_cb(void *ptr, char *str)
{
  printf("%s", str);
}

static void
clear_cb(void *ptr)
{
  tickbuf[0] = 0;
}

static void
pushback_cb(void *ptr, int attr, char *str)
{
  strncat(tickbuf, str, BUFSIZE - 1);
}

static void
update_cb(void *ptr)
{
  char tmp[BUFSIZE * 2];
  sprintf(tmp, "[%s]:%s\n", modebuf, tickbuf);
#if DEBUG
  printf("%s", tmp);
#else 
  udsendto(((struct agent_context *) ptr)->sfd, tmp);
#endif
}

static void
mode_update_cb(void *ptr, int mode)
{
  uim_context uc = ((struct agent_context *) ptr)->uc;
  strncpy(modebuf, uim_get_mode_name(uc, mode), BUFSIZE-1);
}

static void
setmode(uim_context uc)
{
  strncpy(modebuf, uim_get_mode_name(uc, uim_get_current_mode(uc)), BUFSIZE-1);
}

static void
init_agent(int sfd)
{
  uim_context uc;
  int i, nr;
  if (uim_init() == -1) {
    printf("failed to init\n");
    exit(0);
  }
  /**/
  uc =uim_create_context(&default_context,
			 "EUC-JP", NULL, "skk",
			 commit_cb);
  nr = uim_get_nr_im(uc);
  uim_set_preedit_cb(uc, clear_cb, pushback_cb, update_cb);
  setmode(uc);
  uim_set_mode_cb(uc, mode_update_cb);
  default_context.uc = uc;
  default_context.sfd = sfd;
  default_context.next = NULL;
}

static int
needraw(int uc)
{
  if (uc >= 0x20 && uc <= 0x7e)
    return 0;
  switch (uc) {
  case 0x08: /* BS */
  case 0x0a: /* C-j */
  case 0x0d: /* CR */
  case 0x1b: /* ESC */
    return 0;
  }
  return 1;
}

static int
tty2key(int uc)
{
  if (uc >= 0x41 && uc <= 0x5a)
    return uc + ('a' - 'A');
  switch (uc) {
  case 0x08:
    return UKey_Backspace;
  case 0x0a:
    return UKey_Zenkaku_Hankaku;
  case 0x0d:
    return UKey_Return;
  case 0x1b:
    return UKey_Escape;
  }
  return uc;
}

static int
tty2mod(int uc)
{
  if (uc >= 0x41 && uc <= 0x5a)
    return UKey_Shift;
  if (uc >= 0x01 && uc <= 0x1a)
    return UKey_Control;
  return 0;
}

static char *
usersockname()
{
    static char buf[BUFSIZE];
    sprintf(buf, "/tmp/screen-uim-%s", getenv("USER"));
    return buf;
}

int
main(int argc, char **argv)
{
  struct agent_context *ac = &default_context;
  unsigned char uc;
  int sfd;

  if (isatty(0)) {
    settermraw();
    setbuf(stdout, NULL);
  }
  if (argc > 1)
    sockpath = argv[1];
  else
    sockpath = usersockname();
  sfd = mkudclid(sockpath);
  if (sfd < 0)
    return 1;
  init_agent(sfd);

  while (read(0, &uc, 1) == 1) {
    int key, mod, rf;
    if (needraw(uc)) {
	printf("%c", uc);
    } else {
      key = tty2key(uc);
      mod = tty2mod(uc);
      rf = uim_press_key(ac->uc, key, mod);
      uim_release_key(ac->uc, key, mod);
      if (rf)
	printf("%c", uc);
    }
  }
  return 0;
}

void settermraw()
{
  struct termios term;
  if (tcgetattr(0, &term) < 0)
    return;
  memcpy(&tbuf, &term, sizeof(tbuf));

  signal(SIGINT, sighandler);
  signal(SIGQUIT, sighandler);
  signal(SIGTERM, sighandler);
  signal(SIGHUP, sighandler);

  term.c_lflag &= ~(ICANON | ISIG | ECHO | IEXTEN);
  term.c_iflag &= ~(INPCK | ICRNL | ISTRIP | IXON | BRKINT);
  term.c_oflag &= ~OPOST;
  term.c_cflag &= ~(CSIZE | PARENB);
  term.c_cflag |= CS8;
  term.c_cc[VMIN] = 1;
  term.c_cc[VTIME] = 0;
  tcsetattr(0, TCSAFLUSH, &term);
}

void resetterm()
{
  if (tcgetattr(0, &tbuf) < 0)
    return;
}

void sighandler(int s)
{
  resetterm();
  exit(0);
}
